@@ -76,8 +76,15 @@ struct proxier* proxy_create(
 	}
 
 	out->init.buf = xmalloc( sizeof( struct nbd_init_raw ) );
-	out->req.buf  = xmalloc( NBD_MAX_SIZE );
-	out->rsp.buf  = xmalloc( NBD_MAX_SIZE );
+
+	/* Add on the request / reply size to our malloc to accommodate both
+	 * the struct and the data
+	 */
+	out->req.buf  = xmalloc( NBD_MAX_SIZE + NBD_REQUEST_SIZE );
+	out->rsp.buf  = xmalloc( NBD_MAX_SIZE + NBD_REPLY_SIZE );
+
+	log_context = xmalloc( strlen(s_upstream_address) + strlen(s_upstream_port) + 2 );
+	sprintf(log_context, "%s:%s", s_upstream_address, s_upstream_port);
 
 	return out;
 }
@@ -103,7 +110,7 @@ void proxy_destroy( struct proxier* proxy )
 }
 
 /* Shared between our two different connect_to_upstream paths */
-void proxy_finish_connect_to_upstream( struct proxier *proxy, uint64_t size );
+void proxy_finish_connect_to_upstream( struct proxier *proxy, uint64_t size, uint32_t flags );
 
 /* Try to establish a connection to our upstream server. Return 1 on success,
  * 0 on failure. this is a blocking call that returns a non-blocking socket.
@@ -117,12 +124,13 @@ int proxy_connect_to_upstream( struct proxier* proxy )
 
 	int fd = socket_connect( &proxy->connect_to.generic, connect_from );
 	uint64_t size = 0;
+	uint32_t flags = 0;
 
 	if ( -1 == fd ) {
 		return 0;
 	}
 
-	if( !socket_nbd_read_hello( fd, &size ) ) {
+	if( !socket_nbd_read_hello( fd, &size, &flags ) ) {
 		WARN_IF_NEGATIVE(
 			sock_try_close( fd ),
 			"Couldn't close() after failed read of NBD hello on fd %i", fd
@@ -132,7 +140,7 @@ int proxy_connect_to_upstream( struct proxier* proxy )
 
 	proxy->upstream_fd = fd;
 	sock_set_nonblock( fd, 1 );
-	proxy_finish_connect_to_upstream( proxy, size );
+	proxy_finish_connect_to_upstream( proxy, size, flags );
 
 	return 1;
 }
@@ -188,7 +196,7 @@ error:
 	return;
 }
 
-void proxy_finish_connect_to_upstream( struct proxier *proxy, uint64_t size ) {
+void proxy_finish_connect_to_upstream( struct proxier *proxy, uint64_t size, uint32_t flags ) {
 
 	if ( proxy->upstream_size == 0 ) {
 		info( "Size of upstream image is %"PRIu64" bytes", size );
@@ -201,6 +209,17 @@ void proxy_finish_connect_to_upstream( struct proxier *proxy, uint64_t size ) {
 
 	proxy->upstream_size = size;
 
+	if ( proxy->upstream_flags == 0 ) {
+		info( "Upstream transmission flags set to %"PRIu32"", flags );
+	} else if ( proxy->upstream_flags != flags ) {
+		warn(
+			"Upstream transmission flags changed from %"PRIu32" to %"PRIu32"",
+			proxy->upstream_flags, flags
+		);
+	}
+
+	proxy->upstream_flags = flags;
+
 	if ( AF_UNIX != proxy->connect_to.family ) {
 		if ( sock_set_tcp_nodelay( proxy->upstream_fd, 1 ) == -1 ) {
 			warn( SHOW_ERRNO( "Failed to set TCP_NODELAY" ) );
@@ -302,7 +321,7 @@ int proxy_prefetch_for_request( struct proxier* proxy, int state )
 	struct nbd_request_raw* req_raw = (struct nbd_request_raw*) proxy->req.buf;
 	struct nbd_reply_raw *rsp_raw = (struct nbd_reply_raw*) proxy->rsp.buf;
 
-	int is_read = ( req->type & REQUEST_MASK ) == REQUEST_READ;
+	int is_read = req->type == REQUEST_READ;
 
 	if ( is_read ) {
 		/* See if we can respond with what's in our prefetch
@@ -432,20 +451,23 @@ int proxy_read_from_downstream( struct proxier *proxy, int state )
 	if ( proxy->req.needle == NBD_REQUEST_SIZE ) {
 		nbd_r2h_request( request_raw, request );
 
-		if ( ( request->type & REQUEST_MASK ) == REQUEST_DISCONNECT ) {
+		if ( request->type == REQUEST_DISCONNECT ) {
 			info( "Received disconnect request from client" );
 			return EXIT;
 		}
 
-		/* Simple validations */
-		if ( ( request->type & REQUEST_MASK ) == REQUEST_READ ) {
-			if (request->len > ( NBD_MAX_SIZE - NBD_REPLY_SIZE ) ) {
+		/* Simple validations -- the request / reply size have already
+		 * been taken into account in the xmalloc, so no need to worry
+		 * about them here
+		 */
+		if ( request->type == REQUEST_READ ) {
+			if ( request->len > NBD_MAX_SIZE ) {
 				warn( "NBD read request size %"PRIu32" too large", request->len );
 				return EXIT;
 			}
 		}
-		if ( (request->type & REQUEST_MASK ) == REQUEST_WRITE ) {
-			if (request->len > ( NBD_MAX_SIZE - NBD_REQUEST_SIZE ) ) {
+		if ( request->type == REQUEST_WRITE ) {
+			if ( request->len > NBD_MAX_SIZE ) {
 				warn( "NBD write request size %"PRIu32" too large", request->len );
 				return EXIT;
 			}
@@ -456,8 +478,8 @@ int proxy_read_from_downstream( struct proxier *proxy, int state )
 
 	if ( proxy->req.needle == proxy->req.size ) {
 		debug(
-			"Received NBD request from downstream. type=%"PRIu32" from=%"PRIu64" len=%"PRIu32,
-			request->type, request->from, request->len
+			"Received NBD request from downstream. type=%"PRIu16" flags=%"PRIu16" from=%"PRIu64" len=%"PRIu32,
+			request->type, request->flags, request->from, request->len
 		);
 
 		/* Finished reading, so advance state. Leave size untouched so the next
@@ -513,11 +535,15 @@ int proxy_read_init_from_upstream( struct proxier* proxy, int state )
 
 	if ( proxy->init.needle == proxy->init.size ) {
 		uint64_t upstream_size;
-		if ( !nbd_check_hello( (struct nbd_init_raw*) proxy->init.buf, &upstream_size ) ) {
+		uint32_t upstream_flags;
+		if ( !nbd_check_hello( (struct nbd_init_raw*) proxy->init.buf, &upstream_size, &upstream_flags ) ) {
 			warn( "Upstream sent invalid init" );
 			goto disconnect;
 		}
 
+		/* record the flags, and log the reconnection, set TCP_NODELAY */
+		proxy_finish_connect_to_upstream( proxy, upstream_size, upstream_flags );
+
 		/* Currently, we only get disconnected from upstream (so needing to come
 		 * here) when we have an outstanding request. If that becomes false,
 		 * we'll need to choose the right state to return to here */
@@ -599,12 +625,7 @@ int proxy_read_from_upstream( struct proxier* proxy, int state )
 			goto disconnect;
 		}
 
-		if ( reply->error != 0 ) {
-			warn( "NBD error returned from upstream: %"PRIu32, reply->error );
-			goto disconnect;
-		}
-
-		if ( ( proxy->req_hdr.type & REQUEST_MASK ) == REQUEST_READ ) {
+		if ( proxy->req_hdr.type == REQUEST_READ ) {
 			/* Get the read reply data too. */
 			proxy->rsp.size += proxy->req_hdr.len;
 		}
@@ -680,7 +701,7 @@ void proxy_session( struct proxier* proxy )
 
 
 	/* First action: Write hello to downstream */
-	nbd_hello_to_buf( (struct nbd_init_raw *) proxy->rsp.buf, proxy->upstream_size );
+	nbd_hello_to_buf( (struct nbd_init_raw *) proxy->rsp.buf, proxy->upstream_size, proxy->upstream_flags );
 	proxy->rsp.size = sizeof( struct nbd_init_raw );
 	proxy->rsp.needle = 0;
 	state = WRITE_TO_DOWNSTREAM;
