@@ -78,6 +78,8 @@ struct server * server_create (
 	NULLCHECK( out->close_signal );
 	NULLCHECK( out->acl_updated_signal );
 
+	log_context = s_file;
+
 	return out;
 }
 
@@ -330,6 +332,8 @@ int server_count_clients( struct server *params )
 {
 	NULLCHECK( params );
 	int i, count = 0;
+	
+	cleanup_client_threads( params->nbd_client, params->max_nbd_clients );
 
 	for ( i = 0 ; i < params->max_nbd_clients ; i++ ) {
 		if ( params->nbd_client[i].thread != 0 ) {
@@ -422,6 +426,9 @@ void accept_nbd_client(
 	int slot;
 	char s_client_address[64] = {0};
 
+	FATAL_IF_NEGATIVE( sock_set_keepalive_params( client_fd, CLIENT_KEEPALIVE_TIME, CLIENT_KEEPALIVE_INTVL, CLIENT_KEEPALIVE_PROBES),
+		"Error setting keepalive parameters on client socket fd %d", client_fd );
+
 
 	if ( !server_should_accept_client( params, client_address, s_client_address, 64 ) ) {
 		FATAL_IF_NEGATIVE( close( client_fd ),
@@ -705,6 +712,15 @@ void serve_init_allocation_map(struct server* params)
 
 	FATAL_IF_NEGATIVE(fd, "Couldn't open %s", params->filename );
 	size = lseek64( fd, 0, SEEK_END );
+
+	/* If discs are not in multiples of 512, then odd things happen,
+	 * resulting in reads/writes past the ends of files.
+	 */ 
+	if ( size != (size & ~0x1ff)) {
+		warn("file does not fit into 512-byte sectors; the end of the file will be ignored.");
+		size &= ~0x1ff;
+	}
+
 	params->size = size;
 	FATAL_IF_NEGATIVE( size, "Couldn't find size of %s",
 			   params->filename );
